@@ -994,10 +994,14 @@ gboolean farsight_stream_set_sending (FarsightStream *self, gboolean sending)
   return FALSE; // this shouldn't happen
 }
 
-gchar *error_messages[FARSIGHT_STREAM_LAST_ERROR + 1] =
+gchar *error_messages[FARSIGHT_STREAM_ERROR_LAST + 1] =
 {
   gettext_noop ("The stream was terminated"),
-  gettext_noop ("An unknown error has occured in the stream")
+  gettext_noop ("An unknown error has occured in the stream"),
+  gettext_noop ("Timed out on attempt to establish a connection"),
+  gettext_noop ("A network error has occured in the stream"),
+  gettext_noop ("An error has occured while setting up the pipeline"),
+  gettext_noop ("An error has occured on a required resource"),
 };
 
 /**
