@@ -46,6 +46,10 @@
 #include <gnet.h>
 #endif
 
+#ifdef HAVE_CLINKC
+#include "helpers/upnpigd.h"
+#endif
+
 #include <gst/gst.h>
 
 #define DEFAULT_AUDIO_PORT 7078
@@ -98,6 +102,8 @@ struct _FarsightRTPStreamPrivate
   GstElement *sink;
   GstCaps *sink_filter;
 
+  gboolean sending;
+
   guint bus_watch;
 
   gint recv_codec_id;
@@ -130,6 +136,8 @@ struct _FarsightRTPStreamPrivate
   guint16 base_local_port;
   guint16 local_port;
   gint local_socket; /* socket used for symetric RTP */
+
+  gboolean use_upnp;
 };
 
 #define FARSIGHT_RTP_STREAM_GET_PRIVATE(o)  \
@@ -202,6 +210,11 @@ static void farsight_rtp_stream_socket_state_changed (gpointer stream,
                                                       gint state);
 static void farsight_rtp_stream_network_error (gpointer stream);
 #endif
+
+static void farsight_rtp_stream_upnp_send_request(FarsightRTPStream *self,
+    FarsightTransportInfo *derived_trans);
+static void farsight_rtp_stream_upnp_close_ports (FarsightRTPStream *self);
+
 static void farsight_rtp_stream_prepare_transports (FarsightStream *self);
 static gboolean farsight_rtp_stream_add_remote_candidate_to_rtpbin (FarsightRTPStream *self, 
                                                                     const gchar *remote_candidate_id);
@@ -354,6 +367,16 @@ farsight_rtp_stream_init (FarsightRTPStream *self)
   self->priv->sink = NULL;
   self->priv->sink_filter = NULL;
 
+  if (farsight_stream_get_direction (FARSIGHT_STREAM (self)) ==
+      FARSIGHT_STREAM_DIRECTION_RECEIVEONLY)
+  {
+    self->priv->sending = FALSE;
+  }
+  else
+  {
+    self->priv->sending = TRUE;
+  }
+
   self->priv->prepared = FALSE;
 
   self->priv->bus_watch = 0;
@@ -378,6 +401,20 @@ farsight_rtp_stream_init (FarsightRTPStream *self)
 
   self->priv->timeout_src = -1;
   self->priv->conn_timeout = DEFAULT_CONN_TIMEOUT;
+
+#ifdef HAVE_CLINKC
+  if (!getenv ("FARSIGHT_DISABLE_UPNP"))
+  {
+    self->priv->use_upnp = TRUE;
+    g_message ("Setting uPnP to true");
+    if (!upnp_cp_init())
+    {
+      g_warning ("Error setting up uPnP");
+      self->priv->use_upnp = FALSE;
+    }
+  }
+#endif
+
 }
 
 static GObject *
@@ -607,8 +644,16 @@ farsight_rtp_stream_bus_watch_cb (GstBus *bus, GstMessage *message,
                  __FUNCTION__, __LINE__, err->code, err->message);
       g_error_free (err);
 
-      farsight_stream_signal_error (stream,
-                                    FARSIGHT_STREAM_UNKNOWN_ERROR, NULL);
+      if (err->domain == GST_RESOURCE_ERROR)
+      {
+        farsight_stream_signal_error (stream,
+            FARSIGHT_STREAM_ERROR_RESOURCE, err->message);
+      }
+      else
+      {
+        farsight_stream_signal_error (stream,
+            FARSIGHT_STREAM_ERROR_UNKNOWN, err->message);
+      }
       farsight_rtp_stream_stop (stream);
       break;
       }
@@ -636,7 +681,7 @@ farsight_rtp_stream_new_payload_type (GstElement *element, gint pt,
         __LINE__, pt);
     farsight_rtp_stream_stop (FARSIGHT_STREAM(stream));
     farsight_stream_signal_error (FARSIGHT_STREAM(stream), 
-                                  FARSIGHT_STREAM_UNKNOWN_ERROR, NULL);
+                                  FARSIGHT_STREAM_ERROR_UNKNOWN, NULL);
     return;
   }
   */
@@ -659,6 +704,18 @@ farsight_rtp_stream_payload_type_change (GstElement *element, gint pt,
   farsight_rtp_stream_create_new_pt_recv_pipeline(self, pt);
 }
 
+static gboolean
+error_on_create_recv_pipeline (gpointer data)
+{
+  FarsightStream *stream = FARSIGHT_STREAM (data);
+
+  farsight_rtp_stream_stop (stream);
+  farsight_stream_signal_error (stream, FARSIGHT_STREAM_ERROR_PIPELINE_SETUP,
+      "Error creating new recv pipeline");
+
+  return FALSE;
+}
+
 static void
 farsight_rtp_stream_create_new_pt_recv_pipeline (FarsightStream *stream, gint id)
 {
@@ -668,6 +725,8 @@ farsight_rtp_stream_create_new_pt_recv_pipeline (FarsightStream *stream, gint id
   GstElement *codec_bin = NULL;
   CodecInternal *codec_internal = NULL;
 
+  /* WARNING This function gets called in the recv pipeline thread */
+
   codec_internal = lookup_codec_by_pt (farsight_stream_get_media_type (stream),
       id);
   if (!codec_internal) {
@@ -698,6 +757,8 @@ farsight_rtp_stream_create_new_pt_recv_pipeline (FarsightStream *stream, gint id
     }
     gst_bin_add (GST_BIN (self->priv->pipeline), codec_bin);
 
+    self->priv->recv_codec_id = id;
+
     /* we emit now so we can set anything we need on the elements before they
      * are playing */
     farsight_stream_signal_codec_changed (stream, id);
@@ -728,7 +789,9 @@ farsight_rtp_stream_create_new_pt_recv_pipeline (FarsightStream *stream, gint id
         g_debug ("%s (%d): linking audio sink to codec bin",
             __FUNCTION__, __LINE__);
 
-        if (!self->priv->main_pipeline)
+        if ((self->priv->main_pipeline && 
+              gst_element_get_parent (self->priv->sink) == NULL)
+            || !self->priv->main_pipeline)
         {
           gst_bin_add (GST_BIN (self->priv->pipeline), self->priv->sink);
         }
@@ -751,7 +814,7 @@ farsight_rtp_stream_create_new_pt_recv_pipeline (FarsightStream *stream, gint id
 
         gst_object_unref (GST_OBJECT (old_codec_bin));
       }
-      gst_element_set_state (self->priv->sink, GST_STATE_PAUSED);
+      gst_element_set_state (self->priv->sink, GST_STATE_READY);
       gst_element_link_filtered (GST_ELEMENT (codec_bin), self->priv->sink, self->priv->sink_filter);
       gst_element_set_state (self->priv->sink, GST_STATE_PLAYING);
 
@@ -765,13 +828,9 @@ farsight_rtp_stream_create_new_pt_recv_pipeline (FarsightStream *stream, gint id
 
       gst_pad_set_blocked_async (codec_bin_src_pad, TRUE,
           blocked_cb, NULL);
-
-      gst_object_unref (GST_OBJECT (codec_bin_src_pad));
     }
   }
 
-  self->priv->recv_codec_id = id;
-
   /* restart self pipeline */
   //gst_xml_write_file (GST_ELEMENT (self->priv->pipeline), stdout);
   g_debug ("%s: setting state to playing", G_STRFUNC);
@@ -779,9 +838,7 @@ farsight_rtp_stream_create_new_pt_recv_pipeline (FarsightStream *stream, gint id
 
 error:
   g_warning ("%s: PT change failed", G_STRFUNC);
-  farsight_rtp_stream_stop (stream);
-  farsight_stream_signal_error (stream, FARSIGHT_STREAM_UNKNOWN_ERROR, 
-      "Error creating new recv pipeline");
+  g_idle_add (error_on_create_recv_pipeline, stream);
 }
 
 static GstElement*
@@ -822,7 +879,9 @@ farsight_rtp_stream_set_active_codec (FarsightStream *stream, gint id)
    * used sends a new_segment after a reconnection, this is done by
    * overwriting the link/unlink functions
    * in the current case without the new segment the timestamp stops
-   * incrementing */
+   * incrementing 
+   * Also make sure that if there is a user defined pipeline, we change the
+   * ghost pad target to the new send_codec_bin */
 
   FarsightRTPStream *self = FARSIGHT_RTP_STREAM (stream);
   gchar *name;
@@ -834,6 +893,9 @@ farsight_rtp_stream_set_active_codec (FarsightStream *stream, gint id)
   g_debug ("%s (%d): called to change codec from %d to %d", __FUNCTION__,
           __LINE__, farsight_rtp_stream_get_active_codec(stream), id);
 
+  g_debug ("%s: this does not work yet, returning", __FUNCTION__);
+  return;
+
   if (id == farsight_rtp_stream_get_active_codec(stream))
     return;
 
@@ -963,7 +1025,7 @@ set_var:
 
 error:
   farsight_rtp_stream_stop (stream);
-  farsight_stream_signal_error (stream, FARSIGHT_STREAM_UNKNOWN_ERROR, 
+  farsight_stream_signal_error (stream, FARSIGHT_STREAM_ERROR_PIPELINE_SETUP,
       "Error while changing send codec");
 }
 
@@ -1072,35 +1134,49 @@ static void
 unlink_and_replace_sink (GstPad *element1_src_pad, gboolean blocked,
     gpointer user_data)
 {
-  replace_sink_data *data = (replace_sink_data*) user_data;
+  replace_sink_data *data = (replace_sink_data *)user_data;
+  FarsightRTPStream *self = data->self;
 
   unlink_and_replace (element1_src_pad, blocked, data->sink);
 
   /* remove old sink and update references */
-  if (gst_element_set_state (data->self->priv->sink, GST_STATE_NULL) ==
+  if (gst_element_set_state (self->priv->sink, GST_STATE_NULL) ==
       GST_STATE_CHANGE_ASYNC)
   {
     g_debug ("waiting for state change");
-    gst_element_get_state (data->self->priv->sink, NULL, NULL, GST_CLOCK_TIME_NONE);
+    gst_element_get_state (self->priv->sink, NULL, NULL, GST_CLOCK_TIME_NONE);
     g_debug ("done");
   }
-  if (!data->self->priv->main_pipeline)
+  if (!self->priv->main_pipeline)
   {
-    gst_bin_remove (GST_BIN (data->self->priv->pipeline), data->self->priv->sink);
+    gst_bin_remove (GST_BIN (self->priv->pipeline), self->priv->sink);
   }
-  data->self->priv->sink = data->sink;
-  g_free (user_data);
+  /* unref previous sink */
+  gst_object_unref (self->priv->sink);
+  /* set new sink */
+  self->priv->sink = data->sink;
+  g_free (data);
 }
 
 static void
 farsight_rtp_stream_unlink_source (FarsightRTPStream *self)
 {
-  /* we want to safely unlink the source here and maybe replace the downstream
-   * element (send_codec_bin) */
+  /* we want to safely unlink the source here */
   if (!self->priv->src || !self->priv->send_codec_bin)
   {
     return;
   }
+  /* FIXME Decide how to properly unlink the source. If the source is a tee
+   * we don't need to block it, we just unlink. But if the source is something
+   * else, and we just unlink it, it might return a GST_FLOW_ERROR and stop the
+   * whole pipeline. Problem with blocking the source is that blocking is done
+   * async, and by the time we block+unlink the whole farsight stream might have
+   * been unreffed. Doing a sync block might hang forever if the source stops
+   * emitting buffers... For now I just unlink if there is a user defined
+   * pipeline assuming it will be a tee. If there is no user defined pipeline
+   * then there is no need to unlink the source since it will be NULLED along
+   * the whole pipeline */
+#if 0
   GstPad *codec_bin_sink_pad = gst_element_get_pad (self->priv->send_codec_bin, "sink");
   if (codec_bin_sink_pad == NULL)
   {
@@ -1124,6 +1200,12 @@ farsight_rtp_stream_unlink_source (FarsightRTPStream *self)
   {
     g_debug ("Waiting for source src pad to block");
   }
+#endif
+  if (self->priv->main_pipeline)
+  {
+    gst_element_unlink (self->priv->src, self->priv->pipeline);
+  }
+  /* gst_object_unref (GST_OBJECT (codec_bin_sink_pad)); */
 }
 
 static G_CONST_RETURN GList *
@@ -1177,7 +1259,7 @@ farsight_rtp_stream_set_remote_codecs (FarsightStream *stream,
   else
   {
     farsight_stream_signal_error (FARSIGHT_STREAM(self),
-        FARSIGHT_STREAM_UNKNOWN_ERROR, 
+        FARSIGHT_STREAM_ERROR_UNKNOWN, 
         "You need to run farsight_stream_prepare_transports() before setting"
         "the remote candidates");
     return;
@@ -1344,15 +1426,20 @@ farsight_rtp_stream_set_sink (FarsightStream *stream, GstElement *sink)
   {
     GstPad *codec_bin_src_pad =
       gst_element_get_pad (codec_bin, "src");
-    gst_object_ref (GST_OBJECT (codec_bin_src_pad));
 
-    /* in this case we just want to block the recv bin */
+    /* no new sink */
     if (sink == NULL)
     {
-      gst_pad_set_blocked_async (codec_bin_src_pad, TRUE, blocked_cb,
+      /* in this case we just want to block the recv bin and unlink the sink */
+      replace_sink_data *user_data = g_new0 (replace_sink_data, 1);
+      user_data->self = self;
+      user_data->sink = NULL;
+
+      gst_pad_set_blocked_async (codec_bin_src_pad, TRUE, unlink_and_replace_sink,
           NULL);
       self->priv->sink = NULL;
     }
+    /* new sink provided */
     else
     {
       if (!self->priv->main_pipeline)
@@ -1363,10 +1450,13 @@ farsight_rtp_stream_set_sink (FarsightStream *stream, GstElement *sink)
        * sink */
       if (self->priv->sink == NULL)
       {
+        gst_element_set_state (sink, GST_STATE_READY);
         gst_element_link (codec_bin, sink);
         gst_element_set_state (sink, GST_STATE_PLAYING);
         gst_pad_set_blocked_async (codec_bin_src_pad, FALSE, unblocked_cb,
           NULL);
+        gst_object_ref (sink);
+        self->priv->sink = sink;
       }
       else
       {
@@ -1380,10 +1470,12 @@ farsight_rtp_stream_set_sink (FarsightStream *stream, GstElement *sink)
               unlink_and_replace_sink, user_data))
         {
           g_debug ("Recv pipeline already blocked, connecting new sink");
+          gst_element_set_state (sink, GST_STATE_READY);
           gst_element_link (codec_bin, sink);
           gst_element_set_state (sink, GST_STATE_PLAYING);
           gst_pad_set_blocked_async (codec_bin_src_pad, FALSE, unblocked_cb,
               NULL);
+          self->priv->sink = sink;
         }
       }
     }
@@ -1391,7 +1483,10 @@ farsight_rtp_stream_set_sink (FarsightStream *stream, GstElement *sink)
   else
   {
     self->priv->sink = sink;
-    gst_object_ref(sink);
+    if (sink)
+    {
+      gst_object_ref(sink);
+    }
   }
 
   return TRUE;
@@ -1416,11 +1511,17 @@ farsight_rtp_stream_get_sink (FarsightStream *stream)
   if (self->priv->sink)
     return self->priv->sink;
 
+  /* see if we have a sink inside the pipeline */
+  /* we need a pipeline in this case */
+  if (!self->priv->pipeline)
+  {
+    return NULL;
+  }
+
   GstElement *codec_bin = NULL;
   GstIterator *iter;
   gchar *name;
 
-  /* see if we have a sink inside the pipeline */
   /* Let's see if a recv codec bin for this codec already exists */
   name = g_strdup_printf ("recv%d", self->priv->recv_codec_id);
   codec_bin = gst_bin_get_by_name (GST_BIN(self->priv->pipeline), name);
@@ -1490,16 +1591,29 @@ farsight_rtp_stream_build_base_pipeline (FarsightRTPStream *self)
     {
       /* Create a bin and a gst bus for it and set it on the bin */
       self->priv->pipeline = gst_bin_new (NULL);
+      if (!self->priv->pipeline)
+      {
+        goto error;
+      }
       /* let's make our bin independant of the parent state changes */
       gst_element_set_locked_state (self->priv->pipeline, TRUE);
       gst_bin_add (GST_BIN (self->priv->main_pipeline), self->priv->pipeline);
       pipe_bus = gst_bus_new ();
       gst_element_set_bus (GST_ELEMENT_CAST (self->priv->pipeline), pipe_bus);
+
+      /* let's create a ghostpad for our sink */
+      GstPad *ghostpad;
+      ghostpad = gst_ghost_pad_new_no_target ("sink", GST_PAD_SINK);
+      gst_element_add_pad (self->priv->pipeline, ghostpad);
     }
     else
     {
       /* Create a pipeline and listen on it's bus */
       self->priv->pipeline = gst_pipeline_new ("pipeline");
+      if (!self->priv->pipeline)
+      {
+        goto error;
+      }
       pipe_bus = gst_pipeline_get_bus (GST_PIPELINE (self->priv->pipeline));
     }
     self->priv->bus_watch = gst_bus_add_watch (pipe_bus, farsight_rtp_stream_bus_watch_cb, self);
@@ -1584,7 +1698,7 @@ error:
     self->priv->pipeline = NULL;
   }
   farsight_stream_signal_error (FARSIGHT_STREAM(self),
-          FARSIGHT_STREAM_UNKNOWN_ERROR, 
+          FARSIGHT_STREAM_ERROR_PIPELINE_SETUP,
           "Error setting up core RTP pipeline");
   return FALSE;
 }
@@ -1596,6 +1710,13 @@ farsight_rtp_stream_build_send_pipeline (FarsightRTPStream *self)
   GstElement *codec_bin;
 
   g_return_val_if_fail (self != NULL, FALSE);
+  /* let's create the base pipeline if not already done so (prepare_transports
+   * not called) */
+  if (!self->priv->rtpbin || !self->priv->pipeline)
+  {
+    farsight_rtp_stream_build_base_pipeline (self);
+  }
+
   g_return_val_if_fail (self->priv->rtpbin != NULL, FALSE);
   g_return_val_if_fail (self->priv->pipeline != NULL, FALSE);
 
@@ -1670,19 +1791,34 @@ farsight_rtp_stream_build_send_pipeline (FarsightRTPStream *self)
 
   gst_bin_add (GST_BIN (self->priv->pipeline), codec_bin);
 
-  /* connect src to codec_bin */
   if (self->priv->src)
   {
-    gchar *tmp_caps = gst_caps_to_string (self->priv->src_filter);
-    g_debug ("linking src %p to codec bin %p with caps %s", self->priv->src, codec_bin,
-        tmp_caps);
-    g_free (tmp_caps);
-    if (!gst_element_link_filtered (self->priv->src, codec_bin, self->priv->src_filter))
+    if ((GstElement *)gst_element_get_parent (self->priv->src) == self->priv->pipeline)
     {
-      g_warning ("Could not link src to codec bin");
-      goto error;
+      /* connect src to codec_bin */
+      gchar *tmp_caps = gst_caps_to_string (self->priv->src_filter);
+      g_debug ("linking src %p to codec bin %p with caps %s", self->priv->src, codec_bin,
+          tmp_caps);
+      g_free (tmp_caps);
+      if (!gst_element_link_filtered (self->priv->src, codec_bin, self->priv->src_filter))
+      {
+        g_warning ("Could not link src to codec bin");
+        goto error;
+      }
+    }
+    else
+    {
+      /* codec_bin and source have different parents, let's use our ghostpad */
+      GstPad *ghostpad = gst_element_get_pad (self->priv->pipeline, "sink");
+      GstPad *codec_bin_sink_pad = gst_element_get_pad (codec_bin, "sink");
+      gst_ghost_pad_set_target ((GstGhostPad *)ghostpad, codec_bin_sink_pad);
+      gst_element_link_filtered (self->priv->src, self->priv->pipeline,
+          self->priv->src_filter);
+      gst_object_unref (ghostpad);
+      gst_object_unref (codec_bin_sink_pad);
     }
   }
+
   /* connect rtpbin to codec_bin */
   if (!gst_element_link_pads (codec_bin, "src", self->priv->rtpbin, "sink%d"))
   {
@@ -1712,6 +1848,13 @@ farsight_rtp_stream_build_send_pipeline (FarsightRTPStream *self)
 
   self->priv->build_send_pipeline = FALSE;
 
+  /* if user requested to block the pad, let's do it */
+  if (self->priv->sending == FALSE)
+  {
+    self->priv->sending = TRUE;
+    farsight_rtp_stream_set_sending (FARSIGHT_STREAM (self), FALSE);
+  }
+
   if (self->priv->start_me)
   {
     farsight_rtp_stream_start (FARSIGHT_STREAM(self));
@@ -1845,6 +1988,14 @@ farsight_rtp_stream_stop (FarsightStream *stream)
   }
 #endif
 
+  if (self->priv->use_upnp)
+  {
+    farsight_rtp_stream_upnp_close_ports (self);
+#ifdef HAVE_CLINKC
+    upnp_cp_shutdown();
+#endif
+  }
+
   farsight_stream_signal_state_changed (stream,
       FARSIGHT_STREAM_STATE_DISCONNECTED,
       farsight_stream_get_direction (stream));
@@ -1855,44 +2006,50 @@ farsight_rtp_stream_set_sending (FarsightStream *stream, gboolean sending)
 {
   FarsightRTPStream *self = (FarsightRTPStream *) stream;
 
-  GstElement *source = farsight_rtp_stream_get_source (stream);
-  GstPad *source_pad = gst_element_get_pad (source, "src");
-  if (!source_pad)
-  {
-    GstElement *codec_bin = NULL;
-    gchar *name = NULL;
-    name = g_strdup_printf ("send%d", self->priv->send_codec_id);
-    codec_bin = gst_bin_get_by_name (GST_BIN(self->priv->pipeline), name);
-    g_free (name);
+  GstElement *codec_bin = NULL;
+  GstPad *codec_bin_src_pad  = NULL;
+  gchar *name = NULL;
 
-    if (codec_bin)
+  if (self->priv->sending == sending)
+  {
+    return TRUE;
+  }
+
+  self->priv->sending = sending;
+
+  name = g_strdup_printf ("send%d", self->priv->send_codec_id);
+  codec_bin = gst_bin_get_by_name (GST_BIN(self->priv->pipeline), name);
+  g_free (name);
+
+  if (codec_bin)
+  {
+    codec_bin_src_pad = gst_element_get_pad (codec_bin, "src");
+    if (!codec_bin_src_pad)
     {
-      GstPad *codec_bin_src_pad = gst_element_get_pad (codec_bin, "sink");
-      source_pad = gst_pad_get_peer (codec_bin_src_pad);
+      g_warning ("send codec has no source pad! This shouldn't happen");
+      return FALSE;
     }
   }
-
-  if (!source_pad)
-  {
-    g_debug ("No source found to change sending state");
-    return FALSE;
-  }
-
-  gst_object_ref (GST_OBJECT (source_pad));
-  if (sending)
-  {
-    gst_pad_set_blocked_async (source_pad, FALSE, unblocked_cb, NULL);
-    g_object_set (G_OBJECT (stream), "direction",
-        FARSIGHT_STREAM_DIRECTION_BOTH, NULL);
-  }
   else
   {
-    gst_pad_set_blocked_async (source_pad, TRUE, blocked_cb, NULL);
-    g_object_set (G_OBJECT (stream), "direction",
-        FARSIGHT_STREAM_DIRECTION_RECEIVEONLY, NULL);
+    g_message ("send codec bin not created yet, will set sending later");
+    return TRUE;
   }
 
-  gst_object_unref (GST_OBJECT (source_pad));
+  if (sending)
+  {
+    g_debug ("Setting sending to %d", sending);
+    gst_pad_set_blocked_async (codec_bin_src_pad, FALSE, unblocked_cb, NULL);
+   /* g_object_set (G_OBJECT (stream), "direction",
+        FARSIGHT_STREAM_DIRECTION_BOTH, NULL);*/
+  }
+  else
+  {
+    g_debug ("Setting sending on %d", sending);
+    gst_pad_set_blocked_async (codec_bin_src_pad, TRUE, blocked_cb, NULL);
+    /*g_object_set (G_OBJECT (stream), "direction",
+        FARSIGHT_STREAM_DIRECTION_RECEIVEONLY, NULL);*/
+  }
 
   return TRUE;
 }
@@ -1913,6 +2070,61 @@ farsight_rtp_stream_choose_codec (FarsightRTPStream *self,
   return FALSE;
 }
 
+static void
+farsight_rtp_stream_upnp_send_request(FarsightRTPStream *self,
+    FarsightTransportInfo *derived_trans)
+{
+#ifdef HAVE_CLINKC
+  FarsightTransportInfo *trans = NULL;
+  const GList *lp;
+  GList *candidate_list = self->priv->local_candidates;
+
+
+  g_debug ("Looking for local ip");
+
+  for (lp = candidate_list; lp; lp = g_list_next (lp)) {
+    trans = (FarsightTransportInfo *) lp->data;
+    if (trans->type == FARSIGHT_CANDIDATE_TYPE_LOCAL)
+    {
+      if (trans->ip)
+      {
+        /* let's ignore loopback */
+        if (g_ascii_strcasecmp(trans->ip, "127.0.0.1") != 0)
+        {
+          g_debug ("Found local_ip %s", trans->ip);
+          /* open ports */
+          upnp_open_port_all_igds (trans->ip, derived_trans->port, derived_trans->port,
+              (derived_trans->proto == FARSIGHT_NETWORK_PROTOCOL_UDP)?
+              IPPROTO_UDP:IPPROTO_TCP);
+        }
+      }
+    }
+  }
+#endif
+}
+
+static void
+farsight_rtp_stream_upnp_close_ports (FarsightRTPStream *self)
+{
+#ifdef HAVE_CLINKC
+  FarsightTransportInfo *trans = NULL;
+  const GList *lp;
+  GList *candidate_list = self->priv->local_candidates;
+
+  for (lp = candidate_list; lp; lp = g_list_next (lp)) {
+    trans = (FarsightTransportInfo *) lp->data;
+    if (trans->type == FARSIGHT_CANDIDATE_TYPE_DERIVED)
+    {
+      g_debug ("Found derived ip %s", trans->ip);
+      /* close ports */
+      upnp_close_port_all_igds (trans->port, 
+          (trans->proto == FARSIGHT_NETWORK_PROTOCOL_UDP)?
+            IPPROTO_UDP:IPPROTO_TCP);
+    }
+  }
+#endif
+}
+
 static gboolean 
 farsight_rtp_stream_candidate_exists (FarsightStream *stream, 
                                       const GList *candidate_list, 
@@ -1977,6 +2189,16 @@ farsight_rtp_stream_native_candidate_ready (gpointer stream,
     trans = (FarsightTransportInfo *) candidate_copy->data;
     farsight_stream_signal_new_native_candidate (stream,
             (gchar *)trans->candidate_id);
+
+    if (self->priv->use_upnp)
+    {
+      /* TODO this assumes the local IP is found before the derived IP */
+      /* if not it won't send any request since it needs the local ip */
+      if (trans->type == FARSIGHT_CANDIDATE_TYPE_DERIVED)
+      {
+        farsight_rtp_stream_upnp_send_request (self, trans);
+      }
+    }
   }
 
   // farsight_stream_signal_native_candidates_prepared (stream);
@@ -2035,7 +2257,7 @@ farsight_rtp_stream_network_error (gpointer stream)
   g_warning ("error from jinglep2p socketmanager.");
   farsight_rtp_stream_stop (stream);
   farsight_stream_signal_error (stream,
-          FARSIGHT_STREAM_UNKNOWN_ERROR, 
+          FARSIGHT_STREAM_ERROR_NETWORK, 
           "Network error from jinglep2p socketmanager");
 }
 #endif
@@ -2053,7 +2275,7 @@ farsight_rtp_stream_connection_timed_out (gpointer data)
   {
     g_warning ("Could not establish a connection");
     farsight_stream_signal_error (FARSIGHT_STREAM(self),
-        FARSIGHT_STREAM_UNKNOWN_ERROR, "Could not establish a connection");
+        FARSIGHT_STREAM_ERROR_TIMEOUT, "Could not establish a connection");
     farsight_rtp_stream_stop (FARSIGHT_STREAM(self));
   }
 
@@ -2086,8 +2308,11 @@ farsight_rtp_stream_prepare_transports (FarsightStream *stream)
 
   socketclient_create_socket (self->priv->socket_client, "rtp");
 
+  g_message ("aaaah");
+
   socketclient_start_processing_candidates (self->priv->socket_client);
 
+  g_message ("aaaah");
   farsight_stream_signal_state_changed (stream,
       FARSIGHT_STREAM_STATE_CONNECTING, 
       farsight_stream_get_direction (stream));
