@@ -545,6 +545,7 @@ replace_audio_sink (gpointer data)
 }
 */
 
+/*
 static gboolean
 stop_stream (gpointer data)
 {
@@ -554,7 +555,9 @@ stop_stream (gpointer data)
 
   return FALSE;
 }
+*/
 
+/*
 static gboolean
 set_sending (gpointer data)
 {
@@ -582,6 +585,49 @@ set_sending (gpointer data)
 
   return TRUE;
 }
+*/
+
+typedef struct {
+  GstElement *sink;
+  GstElement *tee;
+} ChangePreviewOf;
+
+#if 0
+static gboolean
+change_preview_window (gpointer data)
+{
+  ChangePreviewOf *prev_struct = (ChangePreviewOf *)data;
+
+  gst_element_set_state (pipeline, GST_STATE_PAUSED);
+
+  gst_element_unlink (prev_struct->tee, prev_struct->sink);
+
+  gst_element_set_state (prev_struct->sink, GST_STATE_NULL);
+  gst_bin_remove (GST_BIN (pipeline), prev_struct->sink);
+
+  GstElement *preview_bin = gst_parse_bin_from_description
+    ("ffmpegcolorspace ! videoscale ! ximagesink sync=FALSE",
+     TRUE, NULL);
+  gst_bin_add (GST_BIN (pipeline), preview_bin);
+
+  gst_element_link (prev_struct->tee, preview_bin);
+
+  gst_element_set_state (pipeline, GST_STATE_PLAYING);
+
+
+  /*
+  ChangePreviewOf *nprev_struct = g_new0 (ChangePreviewOf, 1);
+
+  nprev_struct->tee = prev_struct->tee;
+  nprev_struct->sink = preview_bin;
+
+  g_timeout_add (3000, change_preview_window, nprev_struct);
+  */
+
+  g_free (prev_struct);
+  return FALSE;
+}
+#endif
 
 int
 main (int argc, char **argv)
@@ -659,6 +705,8 @@ main (int argc, char **argv)
     //gst_element_link (alsasrc, tee);
     //gst_element_link_filtered (tee, fakesink, prev_caps);
     //gst_element_link (tee, queue);
+    //
+    gst_bin_add (GST_BIN (pipeline), alsasink);
 
     g_object_set (G_OBJECT (alsasink), "sync", FALSE, NULL);
     g_object_set (G_OBJECT (alsasink), "latency-time", G_GINT64_CONSTANT (20000),
@@ -677,24 +725,34 @@ main (int argc, char **argv)
     GstElement *videosrc, *videoscale, *colorspace,
                *glimagesink;
     GstPad *pad;
+    GstElement *fakesink;
 
     GstCaps *src_filter = gst_caps_new_simple ("video/x-raw-yuv", "width",
         G_TYPE_INT, 352, "height", G_TYPE_INT, 288, "framerate",
         GST_TYPE_FRACTION, 15,1, NULL);
 
+    fakesink = gst_element_factory_make ("fakesink", NULL);
     videosrc = gst_parse_bin_from_description ("videotestsrc is-live=TRUE ! video/x-raw-yuv, width=352, height=288, framerate=15/1 ! timeoverlay", 
         TRUE, NULL);
     tee = gst_element_factory_make ("tee", NULL);
     gst_bin_add_many (GST_BIN (pipeline), videosrc, tee, NULL);
     gst_element_link (videosrc, tee);
 
+    ChangePreviewOf *prev_struct = g_new0 (ChangePreviewOf, 1);
     GstElement *preview_bin = gst_parse_bin_from_description
       ("ffmpegcolorspace ! videoscale ! ximagesink sync=FALSE",
        TRUE, NULL);
     gst_bin_add (GST_BIN (pipeline), preview_bin);
+    gst_bin_add (GST_BIN (pipeline), fakesink);
 
     //gst_element_link_filtered (tee, preview_bin, src_filter);
     gst_element_link (tee, preview_bin);
+    gst_element_link (tee, fakesink);
+
+    prev_struct->sink = preview_bin;
+    prev_struct->tee = tee;
+
+    //g_timeout_add (6000, change_preview_window, prev_struct);
 
     sinktype = getenv ("FS_TEST_VIDEO_SINK");
 
@@ -745,6 +803,7 @@ main (int argc, char **argv)
 
     //farsight_stream_set_source_filter (video_stream, src_filter);
     gst_caps_unref (src_filter);
+    gst_bin_add (GST_BIN (pipeline), videosink);
 
     /* let's connect the the prepare-xwindow-id bus message sent by xvimagesink */
     if (argc == 6)
@@ -760,7 +819,7 @@ main (int argc, char **argv)
 
   gst_element_set_state (pipeline, GST_STATE_PLAYING);
 
-  sleep (3);
+ // sleep (3);
 
   if (mode == MODE_AUDIO || mode == MODE_DUPLEX)
   {
@@ -772,22 +831,24 @@ main (int argc, char **argv)
     video_stream = setup_rtp_stream (session, FARSIGHT_MEDIA_TYPE_VIDEO);
   }
 
+  g_debug ("ok next");
+
   if (audio_stream)
   {
     farsight_stream_set_active_codec (audio_stream, 8);
     farsight_stream_set_source (audio_stream, alsasrc);
     farsight_stream_set_sink (audio_stream, alsasink);
-    g_timeout_add (100000, stop_stream, audio_stream);
+    //g_timeout_add (5000, stop_stream, audio_stream);
   }
   if (video_stream)
   {
     farsight_stream_set_sink(video_stream, videosink);
     farsight_stream_set_source(video_stream, tee);
-    g_timeout_add (100000, stop_stream, video_stream);
+    //g_timeout_add (5000, stop_stream, video_stream);
   }
 
   //g_timeout_add (5000, replace_audio_sink, NULL);
-  g_timeout_add (3000, set_sending, NULL);
+  //g_timeout_add (3000, set_sending, NULL);
   //g_timeout_add (7500, null_audio_sink, NULL);
 
   g_main_loop_run (mainloop);
@@ -846,11 +907,17 @@ setup_rtp_stream (FarsightSession * session, FarsightMediaType type)
         codec->clock_rate);
   }
 
+  g_debug ("wazap");
+
   send_codecs (possible_codecs);
 
+  g_debug ("wazap");
+
   farsight_stream_set_pipeline (stream, (GstElement *)pipeline);
 
+  g_debug ("wazap");
   farsight_stream_prepare_transports (stream);
+  g_debug ("wazap");
 
   return stream;
 }
